import { useRef, useCallback, useState, useEffect, useMemo } from 'react';

class ConcurrencyError extends Error {
    constructor(m) {
        super(m !== null && m !== void 0 ? m : 'Concurrent execution is not permitted');
    }
}

/******************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
/* global Reflect, Promise, SuppressedError, Symbol */


function __rest(s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
}

function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

typeof SuppressedError === "function" ? SuppressedError : function (error, suppressed, message) {
    var e = new Error(message);
    return e.name = "SuppressedError", e.error = error, e.suppressed = suppressed, e;
};

/**
 * Keeps a useRef() handle updated to the most recent value, useful
 * to inject unstable values/callback into memoized functions, example:
 *
 * function useSomething(callback) {
 *     const last = useLatest(callback);
 *
 *     const f = useCallback(() => {
 *         last.current?.();
 *     })
 * }
 *
 */
function useLatest(state) {
    const ref = useRef(state);
    ref.current = state;
    return ref;
}

/**
 * Wraps an async function and returns a plain function to be used inside sync functions
 * and component callbacks, any exceptions will catch and console.debug() for
 * development purposes.
 *
 * (!) Remember to enable Verbose Debug Levels to see this log
 *
 * (?) The returned function is stable
 *
 * /B 2022-10-19
 */
function usePromise(fn) {
    const lastFn = useLatest(fn);
    return useCallback((...args) => {
        lastFn.current.apply(undefined, args).catch((error) => {
            console.debug(`usePromise collected an error ${fn.name}`, error);
        });
    }, []);
}

function useReset(initialState) {
    const [state, setState] = useState(initialState);
    const latestState = useLatest(initialState);
    const resetState = useCallback(() => { setState(latestState.current); }, []);
    return [state, setState, resetState];
}

function useAsync(fn) {
    const lastFn = useLatest(fn);
    // Unique call id to handle concurrency
    const callIdRef = useRef(0);
    // Inner state
    const [state, setState, reset] = useReset({
        isLoading: false,
        isCompleted: false,
        isSucceed: false,
        isFailed: false,
    });
    const runAsync = useCallback((...args) => __awaiter(this, void 0, void 0, function* () {
        const callId = ++callIdRef.current;
        setState({ isLoading: true, isCompleted: false, isSucceed: false, isFailed: false, args });
        try {
            // Execute
            const result = yield lastFn.current(...args);
            if (callIdRef.current === callId) {
                setState({ isLoading: false, isCompleted: true, isSucceed: true, isFailed: false, args, result });
            }
            return result;
        }
        catch (error) {
            if (callIdRef.current === callId) {
                setState({ isLoading: false, isCompleted: true, isSucceed: false, isFailed: true, args, error });
            }
            throw error;
        }
    }), []);
    const run = usePromise(runAsync);
    return Object.assign(Object.assign({}, state), { runAsync,
        run,
        reset });
}

/**
 * Syntactic sugar to have effects with async functions
 *
 * /B 2022-10-19
 */
function useAsyncEffect(fn, deps) {
    const lastFn = useLatest(fn);
    useEffect(() => {
        lastFn.current().catch((error) => {
            console.debug(`useAsyncEffect collected an error ${fn.name}`, error);
        });
    }, deps);
}

/**
 * Wraps an async function such that only one call at-a-time may be in execution,
 * any attempt to call the promise while another execution is in progress
 * will result in the second call throwing an exception.
 *
 * (!) Promise should be stable, changing promise while a call is in progress
 * will not affect the outcome nor interrupt the execution.
 *
 * (?) Returned task is assured to be stable, an old result instance will
 * always call the most fresh task.
 *
 * /B 2022-10-19
 */
function useExclusivePromise(fn) {
    const lastFn = useLatest(fn);
    const isExecuting = useRef(false);
    return useCallback((...args) => __awaiter(this, void 0, void 0, function* () {
        if (isExecuting.current) {
            throw new ConcurrencyError(`Promise is already running`);
        }
        else {
            isExecuting.current = true;
            const res = yield lastFn.current.apply(undefined, args);
            isExecuting.current = false;
            return res;
        }
    }), []);
}

function resolveHookInitialValue(initialValue) {
    return typeof initialValue === 'function' ? initialValue() : initialValue;
}
function resolveHookAction(action, ...args) {
    return typeof action === 'function' ? action(...args) : action;
}
/**
 * Compares two arrays to contain the same values (according to Object.is)
 */
function areArgsEqual(a, b) {
    if (Array.isArray(a) && Array.isArray(b)) {
        for (let i = 0; i < a.length && i < b.length; i++) {
            if (!Object.is(a[i], b[i])) {
                return false;
            }
        }
        return true;
    }
    else {
        return Object.is(a, b);
    }
}

/**
 * Joins multiple promise call into one, only one promise will be running
 * at-a-time, calls are not queued but rather joined. When promises are
 * called with arguments if arguments are the same as the original call
 * the promise will be joined and wait for the completion, if arguments
 * are different the promise will throw a concurrency exception.
 *
 * (?) Arguments are shallow compared with Object.is
 * (same strategy that is used by React on effect dependencies)
 *
 * (!) Task should be stable, changing task while a call is in progress
 * will not affect the outcome nor interrupt the execution.
 *
 * (?) Returned task is assured to be stable, an old result instance will
 * always call the most fresh task.
 *
 * todo : if fn includes states that change a sequence such is:
 *      call() -> setState() -> call() ...... return
 *      may return unexpected values to the the second call.
 *
 * /B 2022-10-19
 */
function useSharedPromise(fn) {
    const lastFn = useLatest(fn);
    const stack = useRef();
    return useCallback((...args) => __awaiter(this, void 0, void 0, function* () {
        return new Promise((res, rej) => {
            var _a;
            if (!stack.current) {
                // First execution
                stack.current = { args, res: [res], rej: [rej] };
                lastFn.current.apply(undefined, args)
                    .then((r) => {
                    var _a;
                    (_a = stack.current) === null || _a === void 0 ? void 0 : _a.res.forEach(f => f(r));
                    stack.current = undefined;
                })
                    .catch((e) => {
                    var _a;
                    (_a = stack.current) === null || _a === void 0 ? void 0 : _a.rej.forEach(f => f(e));
                    stack.current = undefined;
                });
            }
            else {
                // Join or fail
                if (!areArgsEqual((_a = stack.current) === null || _a === void 0 ? void 0 : _a.args, args)) {
                    rej(new ConcurrencyError(`Promise already running with different arguments`));
                }
                else {
                    stack.current.res.push(res);
                    stack.current.rej.push(rej);
                }
            }
        });
    }), []);
}

/**
 * Same to useAsync() but when calling concurrently the task will behave same
 * as useSharedPromise(). Eg.: the call will be joint if the arguments are the
 * same or will throw if arguments have changed.
 */
function useSharedAsync(fn) {
    const _a = useAsync(fn), { runAsync: oldRunAsync, run: oldRun } = _a, rest = __rest(_a, ["runAsync", "run"]);
    const runAsync = useSharedPromise(oldRunAsync);
    const run = usePromise(runAsync);
    return Object.assign(Object.assign({}, rest), { runAsync,
        run });
}

function useQuery(fn, deps) {
    const _a = useSharedAsync(fn), { run, runAsync } = _a, state = __rest(_a, ["run", "runAsync"]);
    useEffect(run, deps);
    return Object.assign(Object.assign({}, state), { reloadAsync: runAsync, reload: run });
}

/**
 * This is a very simple and lightweight solution for mobile browser
 * detection. This solution is meant for non-critical features as no
 * guaranties are given by the methods used for detection. If your
 * application is critical then specialized solutions shall be used,
 * such is https://www.npmjs.com/package/react-device-detect, but
 * you should know that there are 20kB of extra code.
 *
 * This solution is inspired by this answer
 * https://stackoverflow.com/questions/11381673
 * and is deliberately developed with a keen for false-negatives,
 * hence it is likely to have mobile device not detected as such but
 * shall not be possible to have desktop devices detected as mobile.
 */
const UAs = [
    /Android/i,
    /webOS/i,
    /iPhone/i,
    /iPad/i,
    /iPod/i,
    /BlackBerry/i,
    /Windows Phone/i
];
function useIsMobile(options) {
    return useMemo(() => {
        if ((options === null || options === void 0 ? void 0 : options.useExperimental) && navigator.userAgentData) {
            return !!navigator.userAgentData.mobile;
        }
        else {
            return UAs.some(r => navigator.userAgent.match(r));
        }
    }, [navigator.userAgentData, navigator.userAgent]);
}

// export type UseEventOptions<T> = Parameters<AddEventListener<T>>[2];
function useEvent(name, handler, target = window) {
    const latest = useLatest(handler);
    useEffect(() => {
        var _a;
        const _target = target || undefined;
        const _names = Array.isArray(name) ? name : [name];
        const inHandler = (event) => { var _a; return (_a = latest.current) === null || _a === void 0 ? void 0 : _a.call(latest, event); };
        const _on = (_a = _target === null || _target === void 0 ? void 0 : _target.addEventListener) !== null && _a !== void 0 ? _a : _target === null || _target === void 0 ? void 0 : _target.on;
        _on && _names.forEach(name => _on.call(_target, name, inHandler));
        return () => {
            var _a;
            const _off = (_a = _target === null || _target === void 0 ? void 0 : _target.removeEventListener) !== null && _a !== void 0 ? _a : _target === null || _target === void 0 ? void 0 : _target.off;
            _off && _names.forEach(name => _off.call(_target, name, inHandler));
        };
    }, [name, target]);
}

/**
 * A hook that handles creation and destruction of global hidden tags. This hook
 * can be used to create dom elements that are outside the React scope. Used
 * by other hooks that provide macro-behaviours such are download and file select
 *
 * @param tagName tag type
 * @param init tag init function
 * @param options tag creation options
 * @param container tag container (defaults to body)
 */
function useGlobalTag(tagName, init, options, container = document.body) {
    // Ref to the tag
    const ref = useRef();
    useEffect(() => {
        // Create and append to outer body
        ref.current = document.createElement(tagName, options);
        ref.current.style.display = 'none';
        container.appendChild(ref.current);
        init === null || init === void 0 ? void 0 : init(ref.current);
        // Remove the tag upon destruction
        return () => {
            if (!ref.current)
                return;
            container.removeChild(ref.current);
            ref.current = undefined;
        };
    }, []);
    return ref;
}

function useScheduledEffect(fn, at, deps, options) {
    const latestFn = useLatest(fn);
    const hasRun = useRef(false);
    useEffect(() => {
        // Skip
        if (at === undefined || hasRun.current)
            return undefined;
        // Coerce
        const unix = typeof at === 'number' ? at : at.getTime();
        const now = Date.now();
        if (unix <= now) {
            // Run now
            if ((options === null || options === void 0 ? void 0 : options.maximumExpiration) === undefined || (now - unix) < (options === null || options === void 0 ? void 0 : options.maximumExpiration)) {
                latestFn.current();
                hasRun.current = true;
            }
            return undefined;
        }
        else {
            // Schedule for future
            const tm = setTimeout(() => latestFn.current(), unix - now);
            return () => clearTimeout(tm);
        }
    }, [at, ...(deps || [])]);
}

/**
 * File download hook. Returns a function which upon call will initiate a native
 * file download by calling an underlying link object.
 */
function useDownload() {
    const ref = useGlobalTag('a');
    return (url, filename) => {
        if (!ref.current)
            return;
        ref.current.href = url;
        if (filename)
            ref.current.setAttribute('download', filename);
        else
            ref.current.removeAttribute('download');
        ref.current.click();
    };
}

/**
 * File select hook. Returns an async function which upon call will initiate a native
 * file selection modal and resolve upon selection completion.
 *
 * const select = useFileSelect({ multiple: true, accept: ['csv', 'xslx' ] })
 *
 */
function useFileSelect(props) {
    // Create a hidden input field with all the expected configs
    const ref = useGlobalTag('input', el => {
        el.setAttribute('type', 'file');
        (props === null || props === void 0 ? void 0 : props.multiple)
            ? el.setAttribute('multiple', 'true')
            : el.removeAttribute('multiple');
        (props === null || props === void 0 ? void 0 : props.accept) && el.setAttribute('accept', props.accept.join(','));
        (props === null || props === void 0 ? void 0 : props.capture) && el.setAttribute('capture', props.capture);
    });
    return () => {
        return new Promise(resolve => {
            if (!ref.current) {
                resolve(undefined);
            }
            else {
                ref.current.addEventListener('change', (event) => resolve(event.target.files
                    ? Array.from(event.target.files)
                    : undefined), { once: true });
                ref.current.click();
            }
        });
    };
}

/**
 * File select state. Returns a file state and a select function, when called will open
 * the native file selection modal that will affect the files list upon closing.
 */
function useFileSelectState(props) {
    var _a;
    const [files, setFiles] = useState((_a = props === null || props === void 0 ? void 0 : props.initialFiles) !== null && _a !== void 0 ? _a : []);
    const selectFn = useFileSelect(props);
    const selectAsync = useAsync(() => __awaiter(this, void 0, void 0, function* () { var _b; return setFiles((_b = (yield selectFn())) !== null && _b !== void 0 ? _b : []); }));
    return {
        files,
        isSelecting: selectAsync.isLoading,
        select: selectAsync.run,
    };
}

/**
 * Wraps useLatest and adds a useCallback. Produces a stable callback given
 * a possibly unstable input function. The returned function is granted to
 * be stable, but will always call the latest function passed to the hook
 */
function useLatestCallback(fn) {
    const latest = useLatest(fn);
    return useCallback((...args) => latest.current(...args), []);
}

/**
 * This hook will return the previous value of a state. The previous
 * value is the last stable value after the previous rendering cycle.
 */
function usePrevious(state) {
    const ref = useRef();
    // Update ref after the current rendering cycle
    useEffect(() => { ref.current = state; });
    return ref.current;
}

/**
 * Returns an updated version of the window.location object. Using
 * this hook will cause the component to re-render each time location
 * changes.
 *
 * /B 2022-10-19
 */
function useLocation() {
    const [value, _, resetValue] = useReset(() => window.location);
    useEvent(['popstate', 'pushstate', 'replacestate'], resetValue);
    return value;
}

/**
 * Returns the current page url
 *
 * /B 2022-10-19
 */
function useHref() {
    return useLocation().href;
}

/**
 * Extract search param from the query whenever necessary
 *
 * /B 2022-10-19
 */
function useSearchParam(name) {
    const location = useLocation();
    return useMemo(() => new URLSearchParams(location.search).get(name) || undefined, [location]);
}

function useDefault(defaultState, initialState) {
    const [state, setState] = useState(initialState);
    return [
        (state === undefined || state === null) ? defaultState : state,
        setState,
    ];
}

function usePatch(initialState) {
    const [state, setState] = useState(initialState);
    const patchState = useCallback((action) => {
        setState((prevState) => (Object.assign(Object.assign({}, prevState), (resolveHookAction(action, prevState)))));
    }, []);
    return [state, setState, patchState];
}

// Simple implementation of a Storage that will hold the values in ram
class RamStorage {
    get length() {
        return Object.keys(this).length;
    }
    clear() {
        for (const key of Object.keys(this)) {
            delete this[key];
        }
    }
    getItem(key) {
        var _a;
        return (_a = this[key]) !== null && _a !== void 0 ? _a : null;
    }
    key(index) {
        var _a;
        return (_a = Object.keys(this)[index]) !== null && _a !== void 0 ? _a : null;
    }
    removeItem(key) {
        delete this[key];
    }
    setItem(key, value) {
        this[key] = value;
    }
}
const ramStorage = new RamStorage();
function getStorage(type) {
    return type === 'ram' ? ramStorage : type === 'sessionStorage' ? window.sessionStorage : window.localStorage;
}
// Collect all values starting with a prefix and strip them of the prefix
function readStorageMulti(storage, prefix) {
    const res = {};
    let count = 0;
    for (const key of Object.keys(storage)) {
        if (key.startsWith(prefix)) {
            res[key.substring(prefix.length)] = JSON.parse(storage.getItem(key));
            count++;
        }
    }
    return count > 0 ? res : undefined;
}
// Store the entire object with the prefix and delete any other key
function setStorageMulti(storage, prefix, value) {
    // Clear extra keys
    for (const key of Object.keys(storage)) {
        if (key.startsWith(prefix) && (value === null || value === void 0 ? void 0 : value[key.substring(prefix.length)]) === undefined) {
            storage.removeItem(key);
        }
    }
    // Add / update others
    if (value) {
        for (const key of Object.keys(value)) {
            if (value[key] !== undefined) {
                const v = typeof value[key] === 'string' ? value[key] : JSON.stringify(value[key]);
                storage.setItem(prefix + key, v);
            }
        }
    }
}

function useStorageValue(key, initialState, storage = 'localStorage') {
    const [state, dispatch] = useState(() => {
        var _a;
        const v = getStorage(storage).getItem(key);
        return (_a = (v !== null ? JSON.parse(v) : undefined)) !== null && _a !== void 0 ? _a : resolveHookInitialValue(initialState);
    });
    function setState(action) {
        dispatch(old => {
            const nw = resolveHookAction(action, old);
            getStorage(storage).setItem(key, typeof nw === 'string' ? nw : JSON.stringify(nw));
            return nw;
        });
    }
    return [state, setState];
}

function useLocalStorageValue(key, initialState) {
    return useStorageValue(key, initialState, 'localStorage');
}

/**
 * Offers an abstraction layer atop SessionStorage and LocalStorage.
 * This hook expects an objects as its state and stores each key of
 * the object as a JSON ENCODED value.
 *
 * (?) Optionally prefix can be provided to prefix each object key
 *     inside the underlying storage (to prevent collisions)
 *
 * (?) Optionally an initial value can be provided that will be used
 *     when no key for the given prefix will be detected in the store.
 *     Keep in mind that store keys may be anyhow deleted, so never
 *     give for granted their presence (not their pairwise presence)
 *
 * /B 2022-10-19
 */
function useStorage(initialState, prefix = '', storage = 'localStorage') {
    const [state, dispatch] = useState(() => {
        var _a;
        return (_a = readStorageMulti(getStorage(storage), prefix)) !== null && _a !== void 0 ? _a : resolveHookInitialValue(initialState);
    });
    function setState(action) {
        dispatch(old => {
            const nw = resolveHookAction(action, old);
            setStorageMulti(getStorage(storage), prefix, nw);
            return nw;
        });
    }
    // Switch dynamically storage
    const lastStorage = useRef();
    useEffect(() => {
        if (lastStorage.current)
            setStorageMulti(getStorage(lastStorage.current), prefix, undefined);
        lastStorage.current = storage;
        setStorageMulti(getStorage(storage), prefix, state);
    }, [storage]);
    return [state, setState];
}

/**
 * Same as useStorage() but does enable dynamic storage updates by
 * subscribing to the window `storage` event.
 *
 * /B 2022-10-19
 */
function useReactiveStorage(initialState, prefix = '', storage = 'localStorage') {
    const [state, setState] = useStorage(initialState, prefix, storage);
    const ctx = useLatest({ prefix, storage });
    useEvent('storage', event => {
        var _a;
        const c = ctx.current;
        if (((_a = event.key) === null || _a === void 0 ? void 0 : _a.startsWith(c.prefix)) && event.storageArea === getStorage(c.storage)) {
            setState(last => (Object.assign(Object.assign({}, last), { [event.key.substring(c.prefix.length)]: event.newValue })));
        }
    });
    return [state, setState];
}

function useSessionStorageValue(key, initialState) {
    return useStorageValue(key, initialState, 'sessionStorage');
}

/**
 * Simply set an interval by providing a callback and delay, the interval
 * is automatically mounted and unmounted.
 *
 * (?) When delay changes the interval is updated, to disable it just
 *     provide null or undefined as delay.
 *
 * (?) Additionally a list of dependencies can be provided. The dependencies
 *     will cause the interval to be rescheduled (delay is implicitly a
 *     dependency)
 *
 * todo : add enable fn to return a tear down function
 */
function useInterval(fn, ms, deps) {
    const lastFn = useLatest(fn);
    useEffect(() => {
        if (ms === undefined || ms === null)
            return undefined;
        const interval = setInterval(() => lastFn.current(), ms || 0);
        return () => clearInterval(interval);
    }, [ms, ...(deps || [])]);
}

export { ConcurrencyError, areArgsEqual, getStorage, readStorageMulti, resolveHookAction, resolveHookInitialValue, setStorageMulti, useAsync, useAsyncEffect, useDefault, useDownload, useEvent, useExclusivePromise, useFileSelect, useFileSelectState, useGlobalTag, useHref, useInterval, useIsMobile, useLatest, useLatestCallback, useLocalStorageValue, useLocation, usePatch, usePrevious, usePromise, useQuery, useReactiveStorage, useReset, useScheduledEffect, useSearchParam, useSessionStorageValue, useSharedAsync, useSharedPromise, useStorage, useStorageValue };
//# sourceMappingURL=index.js.map
