import { DependencyList, MutableRefObject, Dispatch, SetStateAction } from 'react';

type CallFn = () => void;
type DispatchFn<Args extends any[]> = (...args: Args) => void;
type PromiseFn<Args extends any[], Res> = (...args: Args) => Promise<Res>;
type AsyncFn<Res> = () => Promise<Res>;
type AsyncCallFn = () => Promise<void>;
declare class ConcurrencyError extends Error {
    constructor(m?: string);
}

/**
 * An asynchronous wrapping function, that enables use of async functions within
 * React functional components. This hook handles the entire lifecycle of a promise.
 *
 * (?) When the promise is executing, calling it again will result in the previous
 *     execution being ignored. Eg.: the last call to run / runAsync will always be
 *     the one that is affecting the state.
 *
 * (*) Task will hold a reference to the original calling arguments
 */
interface UseAsyncState<Res, Args extends any[], Err> {
    isLoading: boolean;
    isCompleted: boolean;
    isSucceed: boolean;
    isFailed: boolean;
    args?: Args;
    result?: Res;
    error?: Err;
}
interface UseAsyncReturn<Res = void, Args extends any[] = [], Err = any> extends UseAsyncState<Res, Args, Err> {
    run: (...args: Args) => void;
    runAsync: PromiseFn<Args, Res>;
    reset: () => void;
}
declare function useAsync<Res = void, Args extends any[] = [], Err = any>(fn: PromiseFn<Args, Res>): UseAsyncReturn<Res, Args, Err>;

/**
 * Syntactic sugar to have effects with async functions
 *
 * /B 2022-10-19
 */
declare function useAsyncEffect(fn: AsyncCallFn, deps?: DependencyList): void;

/**
 * Wraps an async function such that only one call at-a-time may be in execution,
 * any attempt to call the promise while another execution is in progress
 * will result in the second call throwing an exception.
 *
 * (!) Promise should be stable, changing promise while a call is in progress
 * will not affect the outcome nor interrupt the execution.
 *
 * (?) Returned task is assured to be stable, an old result instance will
 * always call the most fresh task.
 *
 * /B 2022-10-19
 */
declare function useExclusivePromise<Res, Args extends any[] = []>(fn: PromiseFn<Args, Res>): PromiseFn<Args, Res>;

/**
 * Wraps an async function and returns a plain function to be used inside sync functions
 * and component callbacks, any exceptions will catch and console.debug() for
 * development purposes.
 *
 * (!) Remember to enable Verbose Debug Levels to see this log
 *
 * (?) The returned function is stable
 *
 * /B 2022-10-19
 */
declare function usePromise<Res, Args extends any[] = []>(fn: PromiseFn<Args, Res>): DispatchFn<Args>;

/**
 * An asynchronous wrapping function, that enables use of async functions within
 * React functional components. This hook handles the entire lifecycle of a promise.
 *
 * (?) When the promise is executing, calling it again will result in the previous
 *     execution being ignored. Eg.: the last call to run / runAsync will always be
 *     the one that is affecting the state.
 */
interface UseQueryReturn<Res = void, Err = any> extends UseAsyncState<Res, [], Err> {
    reloadAsync: AsyncFn<Res>;
    reload: CallFn;
}
declare function useQuery<Res = void, Err = any>(fn: AsyncFn<Res>, deps?: DependencyList): UseQueryReturn<Res, Err>;

/**
 * Same to useAsync() but when calling concurrently the task will behave same
 * as useSharedPromise(). Eg.: the call will be joint if the arguments are the
 * same or will throw if arguments have changed.
 */
declare function useSharedAsync<Res = void, Args extends any[] = [], Err = any>(fn: PromiseFn<Args, Res>): UseAsyncReturn<Res, Args, Err>;

/**
 * Joins multiple promise call into one, only one promise will be running
 * at-a-time, calls are not queued but rather joined. When promises are
 * called with arguments if arguments are the same as the original call
 * the promise will be joined and wait for the completion, if arguments
 * are different the promise will throw a concurrency exception.
 *
 * (?) Arguments are shallow compared with Object.is
 * (same strategy that is used by React on effect dependencies)
 *
 * (!) Task should be stable, changing task while a call is in progress
 * will not affect the outcome nor interrupt the execution.
 *
 * (?) Returned task is assured to be stable, an old result instance will
 * always call the most fresh task.
 *
 * todo : if fn includes states that change a sequence such is:
 *      call() -> setState() -> call() ...... return
 *      may return unexpected values to the the second call.
 *
 * /B 2022-10-19
 */
declare function useSharedPromise<Res, Args extends any[] = []>(fn: PromiseFn<Args, Res>): PromiseFn<Args, Res>;

declare global {
    interface Navigator {
        userAgentData: {
            mobile?: boolean;
        };
    }
}
interface UseIsMobileOptions {
    useExperimental?: boolean;
}
declare function useIsMobile(options?: UseIsMobileOptions): boolean;

/**
 * Handle an event listener on the provided target. Inspired by
 * react-use useEvent(), but slightly simplified and removed the
 * requirement for the handler to be a memoized function.
 *
 * This function does also support multiple event names to be
 * provided as an array.
 */
type ListenerTypeFn = (name: string, handler: (event?: any) => void, ...args: any[]) => void;
interface ListenerType1 {
    addEventListener: ListenerTypeFn;
    removeEventListener: ListenerTypeFn;
}
interface ListenerType2 {
    on: ListenerTypeFn;
    off: ListenerTypeFn;
}
type AddEventListener<T> = T extends ListenerType1 ? T['addEventListener'] : T extends ListenerType2 ? T['on'] : never;
type UseEventTarget = ListenerType1 | ListenerType2;
declare function useEvent<T extends UseEventTarget>(name: Parameters<AddEventListener<T>>[0] | Parameters<AddEventListener<T>>[0][], handler?: undefined | null | Parameters<AddEventListener<T>>[1], target?: undefined | null | T | Window): void;

/**
 * A hook that handles creation and destruction of global hidden tags. This hook
 * can be used to create dom elements that are outside the React scope. Used
 * by other hooks that provide macro-behaviours such are download and file select
 *
 * @param tagName tag type
 * @param init tag init function
 * @param options tag creation options
 * @param container tag container (defaults to body)
 */
declare function useGlobalTag<K extends keyof HTMLElementTagNameMap>(tagName: K, init?: (tag: HTMLElementTagNameMap[K]) => void, options?: ElementCreationOptions, container?: HTMLElement): MutableRefObject<HTMLElementTagNameMap[K] | undefined>;

/**
 * This effect will run only once at a specified time (unix or date).
 * If on mount the time is in the past will run immediately or skip the
 * run is `maximumExpiration` option is exceeded.
 */
interface UseScheduledEffectOptions {
    maximumExpiration?: number;
}
declare function useScheduledEffect(fn: () => void, at: number | Date | undefined, deps?: DependencyList, options?: UseScheduledEffectOptions): void;

type DownloadFunction = (url: string, filename?: string) => void;
/**
 * File download hook. Returns a function which upon call will initiate a native
 * file download by calling an underlying link object.
 */
declare function useDownload(): DownloadFunction;

interface UseFileSelectProps {
    multiple?: boolean;
    accept?: string[];
    capture?: 'user' | 'environment';
}
type UseFileSelectFunction = () => Promise<File[] | undefined>;
/**
 * File select hook. Returns an async function which upon call will initiate a native
 * file selection modal and resolve upon selection completion.
 *
 * const select = useFileSelect({ multiple: true, accept: ['csv', 'xslx' ] })
 *
 */
declare function useFileSelect(props?: UseFileSelectProps): UseFileSelectFunction;

interface UseFileSelectStateProps extends UseFileSelectProps {
    initialFiles?: File[];
}
interface UseFileSelectStateReturn {
    files: File[];
    isSelecting: boolean;
    select: () => void;
}
/**
 * File select state. Returns a file state and a select function, when called will open
 * the native file selection modal that will affect the files list upon closing.
 */
declare function useFileSelectState(props?: UseFileSelectStateProps): UseFileSelectStateReturn;

type SimpleFn<Args extends any[], Res> = (...args: Args) => Res;
/**
 * Wraps useLatest and adds a useCallback. Produces a stable callback given
 * a possibly unstable input function. The returned function is granted to
 * be stable, but will always call the latest function passed to the hook
 */
declare function useLatestCallback<Res, Args extends any[] = []>(fn: SimpleFn<Args, Res>): SimpleFn<Args, Res>;

/**
 * Keeps a useRef() handle updated to the most recent value, useful
 * to inject unstable values/callback into memoized functions, example:
 *
 * function useSomething(callback) {
 *     const last = useLatest(callback);
 *
 *     const f = useCallback(() => {
 *         last.current?.();
 *     })
 * }
 *
 */
declare function useLatest<T>(state: T): MutableRefObject<T>;

/**
 * This hook will return the previous value of a state. The previous
 * value is the last stable value after the previous rendering cycle.
 */
declare function usePrevious<T>(state: T): T | undefined;

/**
 * Returns the current page url
 *
 * /B 2022-10-19
 */
declare function useHref(): string | undefined;

/**
 * Returns an updated version of the window.location object. Using
 * this hook will cause the component to re-render each time location
 * changes.
 *
 * /B 2022-10-19
 */
declare function useLocation(): Location;

/**
 * Extract search param from the query whenever necessary
 *
 * /B 2022-10-19
 */
declare function useSearchParam(name: string): string | undefined;

/**
 * A useState() hook that also exposes a function to reset the value
 * back to the default one. If the default value is provided as a function
 * the most recent function instance will be called to generate the value.
 */
declare function useDefault<D>(defaultState: D, initialState: D | (() => D)): [D, Dispatch<SetStateAction<D>>];
declare function useDefault<D = undefined>(defaultState: D): [D | undefined, Dispatch<SetStateAction<D | undefined>>];

type PatchStateAction<D> = Partial<D> | ((prevState: D) => Partial<D>);
/**
 * A state for objects, that supports patching.
 */
declare function usePatch<D>(initialState: D | (() => D)): [D, Dispatch<SetStateAction<D>>, Dispatch<PatchStateAction<D>>];
declare function usePatch<D = undefined>(): [D | undefined, Dispatch<SetStateAction<D | undefined>>, Dispatch<PatchStateAction<D | undefined>>];

type ResetAction = () => void;
/**
 * A useState() hook that also exposes a function to reset the value
 * back to the default one. If the default value is provided as a function
 * the most recent function instance will be called to generate the value.
 */
declare function useReset<D>(initialState: D | (() => D)): [D, Dispatch<SetStateAction<D>>, ResetAction];
declare function useReset<D = undefined>(): [D | undefined, Dispatch<SetStateAction<D | undefined>>, ResetAction];

type StorageTypes = 'sessionStorage' | 'localStorage' | 'ram';
type MultiValue = {
    [K: string]: any;
};
declare function getStorage(type: StorageTypes): Storage;
declare function readStorageMulti(storage: Storage, prefix: string): MultiValue | undefined;
declare function setStorageMulti(storage: Storage, prefix: string, value: MultiValue | undefined): void;

declare function useLocalStorageValue<S>(key: string, initialState: S | (() => S)): [S, Dispatch<SetStateAction<S>>];
declare function useLocalStorageValue<S = undefined>(key: string): [S | undefined, Dispatch<SetStateAction<S | undefined>>];

/**
 * Same as useStorage() but does enable dynamic storage updates by
 * subscribing to the window `storage` event.
 *
 * /B 2022-10-19
 */
declare function useReactiveStorage<S extends MultiValue>(initialState?: S | (() => S), prefix?: string, storage?: StorageTypes): [S, Dispatch<SetStateAction<S>>];

declare function useSessionStorageValue<S>(key: string, initialState: S | (() => S)): [S, Dispatch<SetStateAction<S>>];
declare function useSessionStorageValue<S = undefined>(key: string): [S | undefined, Dispatch<SetStateAction<S | undefined>>];

/**
 * Offers an abstraction layer atop SessionStorage and LocalStorage.
 * This hook expects an objects as its state and stores each key of
 * the object as a JSON ENCODED value.
 *
 * (?) Optionally prefix can be provided to prefix each object key
 *     inside the underlying storage (to prevent collisions)
 *
 * (?) Optionally an initial value can be provided that will be used
 *     when no key for the given prefix will be detected in the store.
 *     Keep in mind that store keys may be anyhow deleted, so never
 *     give for granted their presence (not their pairwise presence)
 *
 * /B 2022-10-19
 */
declare function useStorage<S extends MultiValue>(initialState?: S | (() => S), prefix?: string, storage?: StorageTypes): [S, Dispatch<SetStateAction<S>>];

/**
 * Offers an abstraction layer atop SessionStorage and LocalStorage.
 * This hook is similar to useStorage() but handle a single value.
 *
 * /B 2022-10-19
 */
declare function useStorageValue<S>(key: string, initialState: S | (() => S), storage?: StorageTypes): [S, Dispatch<SetStateAction<S>>];
declare function useStorageValue<S = undefined>(key: string, initialState?: undefined, storage?: StorageTypes): [S | undefined, Dispatch<SetStateAction<S | undefined>>];

/**
 * Simply set an interval by providing a callback and delay, the interval
 * is automatically mounted and unmounted.
 *
 * (?) When delay changes the interval is updated, to disable it just
 *     provide null or undefined as delay.
 *
 * (?) Additionally a list of dependencies can be provided. The dependencies
 *     will cause the interval to be rescheduled (delay is implicitly a
 *     dependency)
 *
 * todo : add enable fn to return a tear down function
 */
declare function useInterval(fn: CallFn, ms?: number | null | undefined, deps?: DependencyList): void;

type HookAction<V, Args extends any[]> = V | ((...args: Args) => V);
declare function resolveHookInitialValue<V>(initialValue?: V | (() => V)): any;
declare function resolveHookAction<V, Args extends any[]>(action: HookAction<V, Args>, ...args: Args): V;
/**
 * Compares two arrays to contain the same values (according to Object.is)
 */
declare function areArgsEqual(a: any[], b: any[]): boolean;

export { type AsyncCallFn, type AsyncFn, type CallFn, ConcurrencyError, type DispatchFn, type DownloadFunction, type HookAction, type ListenerType1, type ListenerType2, type ListenerTypeFn, type MultiValue, type PromiseFn, type ResetAction, type SimpleFn, type StorageTypes, type UseAsyncReturn, type UseAsyncState, type UseEventTarget, type UseFileSelectFunction, type UseFileSelectProps, type UseFileSelectStateProps, type UseFileSelectStateReturn, type UseIsMobileOptions, type UseQueryReturn, type UseScheduledEffectOptions, areArgsEqual, getStorage, readStorageMulti, resolveHookAction, resolveHookInitialValue, setStorageMulti, useAsync, useAsyncEffect, useDefault, useDownload, useEvent, useExclusivePromise, useFileSelect, useFileSelectState, useGlobalTag, useHref, useInterval, useIsMobile, useLatest, useLatestCallback, useLocalStorageValue, useLocation, usePatch, usePrevious, usePromise, useQuery, useReactiveStorage, useReset, useScheduledEffect, useSearchParam, useSessionStorageValue, useSharedAsync, useSharedPromise, useStorage, useStorageValue };
